//
//
//    Copyright 2009-2010 TheSeven
//
//
//    This file is part of the Linux4Nano toolkit.
//
//    TheSeven's iBugger is free software: you can redistribute it and/or
//    modify it under the terms of the GNU General Public License as
//    published by the Free Software Foundation, either version 2 of the
//    License, or (at your option) any later version.
//
//    TheSeven's iBugger is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//    See the GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with the Linux4Nano toolkit.  If not, see <http://www.gnu.org/licenses/>.
//
//



#include <toolkit.h>
#include <nand.h>
#include <ftl.h>



/* Keeps the state of the FTL, both on flash and in memory */
struct ftl_cxt_type
{

    /* Update sequence number of the FTL context, decremented
       every time a new revision of FTL meta data is written. */
    uint32_t usn;

    /* Update sequence number for user data blocks. Incremented
       every time a portion of user pages is written, so that
       a consistency check can determine which copy of a user
       page is the most recent one. */
    uint32_t nextblockusn;

    /* Count of currently free pages in the block pool */
    uint16_t freecount;

    /* Index to the first free block in the blockpool ring buffer */
    uint16_t nextfreeidx;

    /* This is a counter that is used to better distribute block
       wear. It is incremented on every block erase, and if it
       gets too high (300 on writes, 20 on sync), the most and
       least worn block will be swapped (inferring an additional
       block write) and the counter will be decreased by 20. */
    uint16_t swapcounter;

    /* Ring buffer of currently free blocks. nextfreeidx is the
       index to freecount free ones, the other ones are currently
       allocated for scattered page blocks. */
    uint16_t blockpool[0x14];

    /* Alignment to 32 bits */
    uint16_t field_36;

    /* vPages where the block map is stored */
    uint32_t ftl_map_pages[8];

    /* Probably additional map page number space for bigger chips */
    uint8_t field_58[0x28];

    /* vPages where the erase counters are stored */
    uint32_t ftl_erasectr_pages[8];

    /* Seems to be padding */
    uint8_t field_A0[0x70];

    /* Pointer to ftl_map used by Whimory, not used by us */
    uint32_t ftl_map_ptr;

    /* Pointer to ftl_erasectr used by Whimory, not used by us */
    uint32_t ftl_erasectr_ptr;

    /* Pointer to ftl_log used by Whimory, not used by us */
    uint32_t ftl_log_ptr;

    /* Flag used to indicate that some erase counter pages should be committed
       as they were changed more than 100 times since the last commit. */
    uint32_t erasedirty;

    /* Seems to be unused */
    uint16_t field_120;

    /* vBlocks used to store the FTL context, map, and erase
       counter pages. This is also a ring buffer, and the oldest
       page gets swapped with the least used page from the block
       pool ring buffer when a new one is allocated. */
    uint16_t ftlctrlblocks[3];

    /* The last used vPage number from ftlctrlblocks */
    uint32_t ftlctrlpage;

    /* Set on context sync, reset on write, so obviously never
       zero in the context written to the flash */
    uint32_t clean_flag;

    /* Seems to be unused, but gets loaded from flash by Whimory. */
    uint8_t field_130[0x15C];

} __attribute__((packed));


/* Keeps the state of the bank's VFL, both on flash and in memory.
   There is one of these per bank. */
struct ftl_vfl_cxt_type
{

    /* Cross-bank update sequence number, incremented on every VFL
       context commit on any bank. */
    uint32_t usn;

    /* See ftl_cxt.ftlctrlblocks. This is stored to the VFL contexts
       in order to be able to find the most recent FTL context copy
       when mounting the FTL. The VFL context number this will be
       written to on an FTL context commit is chosen semi-randomly. */
    uint16_t ftlctrlblocks[3];

    /* Alignment to 32 bits */
    uint8_t field_A[2];

    /* Decrementing update counter for VFL context commits per bank */
    uint32_t updatecount;

    /* Number of the currently active VFL context block, it's an index
       into vflcxtblocks. */
    uint16_t activecxtblock;

    /* Number of the first free page in the active FTL context block */
    uint16_t nextcxtpage;

    /* Seems to be unused */
    uint8_t field_14[4];

    /* Incremented every time a block erase error leads to a remap,
       but doesn't seem to be read anywhere. */
    uint16_t field_18;

    /* Number of spare blocks used */
    uint16_t spareused;

    /* pBlock number of the first spare block */
    uint16_t firstspare;

    /* Total number of spare blocks */
    uint16_t sparecount;

    /* Block remap table. Contains the vBlock number the n-th spare
       block is used as a replacement for. 0 = unused, 0xFFFF = bad. */
    uint16_t remaptable[0x334];

    /* Bad block table. Each bit represents 8 blocks. 1 = OK, 0 = Bad.
       If the entry is zero, you should look at the remap table to see
       if the block is remapped, and if yes, where the replacement is. */
    uint8_t bbt[0x11A];

    /* pBlock numbers used to store the VFL context. This is a ring
       buffer. On a VFL context write, always 8 pages are written,
       and it passes if at least 4 of them can be read back. */
    uint16_t vflcxtblocks[4];

    /* Blocks scheduled for remapping are stored at the end of the
       remap table. This is the first index used for them. */
    uint16_t scheduledstart;

    /* Probably padding */
    uint8_t field_7AC[0x4C];

    /* First checksum (addition) */
    uint32_t checksum1;

    /* Second checksum (XOR), there is a bug in whimory regarding this. */
    uint32_t checksum2;

} __attribute__((packed));


/* Layout of the spare bytes of each page on the flash */
union ftl_spare_data_type
{

    /* The layout used for actual user data (types 0x40 and 0x41) */
    struct ftl_spare_data_user_type
    {

        /* The lPage, i.e. Sector, number */
        uint32_t lpn;

        /* The update sequence number of that page,
           copied from ftl_cxt.nextblockusn on write */
        uint32_t usn;

        /* Seems to be unused */
        uint8_t field_8;

        /* Type field, 0x40 (data page) or 0x41 (last data page of block) */
        uint8_t type;

        /* ECC mark, usually 0xFF. If an error occurred while reading the
           page during a copying operation earlier, this will be 0x55. */
        uint8_t eccmark;

        /* Seems to be unused */
        uint8_t field_B;

        /* ECC data for the user data */
        uint8_t dataecc[0x28];

        /* ECC data for the first 0xC bytes above */
        uint8_t spareecc[0xC];

    } __attribute__((packed)) user;

    /* The layout used for meta data (other types) */
    struct ftl_spare_data_meta_type
    {

        /* ftl_cxt.usn for FTL stuff, ftl_vfl_cxt.updatecount for VFL stuff */
        uint32_t usn;

        /* Index of the thing inside the page,
           for example number / index of the map or erase counter page */
        uint16_t idx;

        /* Seems to be unused */
        uint8_t field_6;

        /* Seems to be unused */
        uint8_t field_7;

        /* Seems to be unused */
        uint8_t field_8;

       /* Type field:
            0x43: FTL context page
            0x44: Block map page
            0x46: Erase counter page
            0x47: "FTL is currently mounted", i.e. unclean shutdown, mark
            0x80: VFL context page */
        uint8_t type;

        /* ECC mark, usually 0xFF. If an error occurred while reading the
           page during a copying operation earlier, this will be 0x55. */
        uint8_t eccmark;

        /* Seems to be unused */
        uint8_t field_B;

        /* ECC data for the user data */
        uint8_t dataecc[0x28];

        /* ECC data for the first 0xC bytes above */
        uint8_t spareecc[0xC];

    } __attribute__((packed)) meta;

};


/* Pointer to an info structure regarding the flash type used */
const struct nand_device_info_type* ftl_nand_type;

/* Number of banks we detected a chip on */
uint32_t ftl_banks;

/* Block map, used vor pBlock to vBlock mapping */
uint16_t ftl_map[0x2000];

/* VFL context for each bank */
struct ftl_vfl_cxt_type ftl_vfl_cxt[4];

/* FTL context */
struct ftl_cxt_type ftl_cxt;

/* Temporary data buffers for internal use by the FTL */
uint8_t ftl_buffer[0x800] __attribute__((aligned(16)));

/* Temporary spare byte buffer for internal use by the FTL */
union ftl_spare_data_type ftl_sparebuffer __attribute__((aligned(16)));

/* Lowlevel BBT for each bank */
uint8_t ftl_bbt[4][0x410];

/* Erase countes for the vBlocks */
uint16_t ftl_erasectr[0x2000];



/* Finds a device info page for the specified bank and returns its number.
   Used to check if one is present, and to read the lowlevel BBT. */
uint32_t ftl_find_devinfo(uint32_t bank)
{
    /* Scan the last 10% of the flash for device info pages */
    uint32_t lowestBlock = (*ftl_nand_type).blocks
                         - ((*ftl_nand_type).blocks / 10);
    uint32_t block, page, pagenum;
    debug_printf("FTL: Searching for DEVICEINFO on bank %u, lowest block: %u...\n",
                 bank, lowestBlock);
    for (block = (*ftl_nand_type).blocks - 1; block >= lowestBlock; block--)
    {
        page = (*ftl_nand_type).pagesperblock - 8;
        for (; page < (*ftl_nand_type).pagesperblock; page++)
        {
            pagenum = block * (*ftl_nand_type).pagesperblock + page;
            if ((nand_read_page(bank, pagenum, ftl_buffer,
                                &ftl_sparebuffer, 1, 0) & 0x11F) != 0)
                continue;
            if (memcmp(ftl_buffer, "DEVICEINFOSIGN\0", 0x10) == 0)
            {
                debug_printf("FTL: DEVICEINFO for bank %u found at page %u\n",
                             bank, pagenum);
                return pagenum;
            }
        }
    }
    debug_printf("FTL: Failed to find DEVICEINFO for bank %u!\n", bank);
    return 0;
}


/* Checks if all banks have proper device info pages */
uint32_t ftl_has_devinfo(void)
{
    uint32_t i;
    for (i = 0; i < ftl_banks; i++) if (ftl_find_devinfo(i) == 0) return 0;
    return 1;
}


/* Loads the lowlevel BBT for a bank to the specified buffer.
   This is based on some cryptic disassembly and not fully understood yet. */
uint32_t ftl_load_bbt(uint32_t bank, uint8_t* bbt)
{
    uint32_t i, j;
    uint32_t pagebase, page = ftl_find_devinfo(bank), page2;
    uint32_t unk1, unk2, unk3;
    debug_printf("VFL: Trying to load lowlevel BBT on bank %u...\n", bank);
    if (page == 0)
    {
        debug_printf("VFL: Can't load lowlevel BBT for bank %u "
                     "due to missing DEVICEINFO!\n", bank);
        return 1;
    }
    pagebase = page & ~((*ftl_nand_type).pagesperblock - 1);
    debug_printf("VFL: Looking for lowlevel BBT for bank %u at page %u...\n", bank, page);
    if ((nand_read_page(bank, page, ftl_buffer,
                        (uint32_t*)0, 1, 0) & 0x11F) != 0)
    {
        debug_printf("VFL: Read error while loading bank %u lowlevel BBT!\n", bank);
        return 1;
    }
    if (memcmp(&ftl_buffer[0x18], "BBT", 4) != 0)
    {
        debug_printf("VFL: Bad signature on bank %u lowlevel BBT!\n", bank);
        return 1;
    }
    unk1 = ((uint16_t*)ftl_buffer)[0x10];
    unk2 = ((uint16_t*)ftl_buffer)[0x11];
    unk3 = ((uint16_t*)ftl_buffer)[((uint32_t*)ftl_buffer)[4] * 0xC + 10]
         + ((uint16_t*)ftl_buffer)[((uint32_t*)ftl_buffer)[4] * 0xC + 11];
    for (i = 0; i < unk1; i++)
    {
        for (j = 0; ; j++)
        {
            page2 = unk2 + i + unk3 * j;
            if (page2 >= (uint32_t)((*ftl_nand_type).pagesperblock - 8))
                break;
            if ((nand_read_page(bank, pagebase + page2, ftl_buffer,
                                (void*)0, 1, 0) & 0x11F) == 0)
            {
                debug_printf("VFL: Found most recent lowlevel BBT for bank %u"
                             " at page %u\n", bank, pagebase + page2);
                memcpy(bbt, ftl_buffer, 0x410);
                return 0;
            }
        }
    }
    debug_printf("VFL: Can't find most recent bank %u lowlevel BBT!\n", bank);
    return 1;
}


/* Calculates the checksums for the VFL context page of the specified bank */
void ftl_vfl_calculate_checksum(uint32_t bank,
                                uint32_t* checksum1, uint32_t* checksum2)
{
    uint32_t i;
    *checksum1 = 0xAABBCCDD;
    *checksum2 = 0xAABBCCDD;
    for (i = 0; i < 0x1FE; i++)
    {
        *checksum1 += ((uint32_t*)(&ftl_vfl_cxt[bank]))[i];
        *checksum2 ^= ((uint32_t*)(&ftl_vfl_cxt[bank]))[i];
    }
}


/* Checks if the checksums of the VFL context
   of the specified bank are correct */
uint32_t ftl_vfl_verify_checksum(uint32_t bank)
{
    uint32_t checksum1, checksum2;
    ftl_vfl_calculate_checksum(bank, &checksum1, &checksum2);
    if (checksum1 == ftl_vfl_cxt[bank].checksum1) return 0;
    /* The following line is pretty obviously a bug in Whimory,
       but we do it the same way for compatibility. */
    if (checksum2 != ftl_vfl_cxt[bank].checksum2) return 0;
    debug_printf("VFL: Bad VFL CXT checksum!\n");
    return 1;
}


/* Returns a pointer to the most recently updated VFL context,
   used to find out the current FTL context vBlock numbers
   (planetbeing's "maxthing") */
struct ftl_vfl_cxt_type* ftl_vfl_get_newest_cxt(void)
{
    uint32_t i, maxusn;
    struct ftl_vfl_cxt_type* cxt = (struct ftl_vfl_cxt_type*)0;
    maxusn = 0;
    for (i = 0; i < ftl_banks; i++)
        if (ftl_vfl_cxt[i].usn >= maxusn)
        {
            cxt = &ftl_vfl_cxt[i];
            maxusn = ftl_vfl_cxt[i].usn;
        }
    return cxt;
}


/* Checks if the specified pBlock is marked bad in the supplied lowlevel BBT.
   Only used while mounting the VFL. */
uint32_t ftl_is_good_block(uint8_t* bbt, uint32_t block)
{
    if ((bbt[block >> 3] & (1 << (block & 7))) == 0) return 0;
    else return 1;
}


/* Checks if the specified vBlock could be remapped */
uint32_t ftl_vfl_is_good_block(uint32_t bank, uint32_t block)
{
    uint8_t bbtentry = ftl_vfl_cxt[bank].bbt[block >> 6];
    if ((bbtentry & (1 << ((7 - (block >> 3)) & 7))) == 0) return 0;
    else return 1;
}


/* Tries to read a VFL context from the specified bank, pBlock and page */
uint32_t ftl_vfl_read_page(uint32_t bank, uint32_t block,
                           uint32_t startpage, void* databuffer,
                           union ftl_spare_data_type* sparebuffer)
{
    uint32_t i;
    for (i = 0; i < 8; i++)
    {
        uint32_t page = block * (*ftl_nand_type).pagesperblock
                      + startpage + i;
        if ((nand_read_page(bank, page, databuffer,
                            sparebuffer, 1, 1) & 0x11F) == 0)
            if ((*sparebuffer).meta.field_8 == 0
             && (*sparebuffer).meta.type == 0x80)
                return 0;
    }
    return 1;
}


/* Translates a bank and vBlock to a pBlock, following remaps */
uint32_t ftl_vfl_get_physical_block(uint32_t bank, uint32_t block)
{
    if (ftl_vfl_is_good_block(bank, block) == 1) return block;

    uint32_t spareindex;
    uint32_t spareused = ftl_vfl_cxt[bank].spareused;
    for (spareindex = 0; spareindex < spareused; spareindex++)
        if (ftl_vfl_cxt[bank].remaptable[spareindex] == block)
#ifdef VFL_TRACE
            debug_printf("FTL: VFL: Following remapped block: %d => %d\n",
                         block, ftl_vfl_cxt[bank].firstspare + spareindex);
#endif
            return ftl_vfl_cxt[bank].firstspare + spareindex;
    return block;
}


// Reads the specified vPage, dealing with all kinds of trouble
uint32_t ftl_vfl_read(uint32_t vpage, void* buffer, void* sparebuffer,
                      uint32_t checkempty, uint32_t remaponfail)
{
#ifdef VFL_TRACE
    debug_printf("FTL: VFL: Reading page %d\n", vpage);
#endif

    uint32_t ppb = (*ftl_nand_type).pagesperblock * ftl_banks;
    uint32_t syshyperblocks = (*ftl_nand_type).blocks
                            - (*ftl_nand_type).userblocks - 0x17;
    uint32_t abspage = vpage + ppb * syshyperblocks;
    if (abspage >= (*ftl_nand_type).blocks * ppb || abspage < ppb)
    {
        debug_printf("FTL: VFL: vPage %d is out of range!\n", vpage);
        return 4;
    }

    uint32_t bank = abspage % ftl_banks;
    uint32_t block = abspage / ((*ftl_nand_type).pagesperblock * ftl_banks);
    uint32_t page = (abspage / ftl_banks) % (*ftl_nand_type).pagesperblock;
    uint32_t physblock = ftl_vfl_get_physical_block(bank, block);
    uint32_t physpage = physblock * (*ftl_nand_type).pagesperblock + page;

    uint32_t ret = nand_read_page(bank, physpage, buffer,
                                  sparebuffer, 1, checkempty);

    if ((ret & 0x11D) != 0 && (ret & 2) == 0)
    {
        nand_reset(bank);
        ret = nand_read_page(bank, physpage, buffer,
                             sparebuffer, 1, checkempty);
        if (remaponfail == 1 &&(ret & 0x11D) != 0 && (ret & 2) == 0)
            debug_printf("FTL: VFL: Would schedule vBlock %d for remapping!\n", block);
        return ret;
    }

    return ret;
}


/* Mounts the VFL on all banks */
uint32_t ftl_vfl_open(void)
{
    uint32_t i, j, k;
    uint32_t minusn, vflcxtidx, last;
    struct ftl_vfl_cxt_type* cxt;
    uint16_t vflcxtblock[4];

    uint32_t syshyperblocks = (*ftl_nand_type).blocks
                            - (*ftl_nand_type).userblocks - 0x18;

    for (i = 0; i < ftl_banks; i++)
    {
        debug_printf("VFL: Trying to mount VFL on bank %u...\n", i);
        if (!ftl_load_bbt(i, ftl_bbt[i]))
        {
            debug_printf("VFL: Got lowlevel BBT on bank %u, searching for CXT...\n", i);
            for (j = 1; j <= syshyperblocks; j++)
                if (ftl_is_good_block(ftl_bbt[i], j) != 0)
                    if (ftl_vfl_read_page(i, j, 0, ftl_buffer,
                                          &ftl_sparebuffer) == 0)
                    {
                        struct ftl_vfl_cxt_type* cxt;
                        cxt = (struct ftl_vfl_cxt_type*)ftl_buffer;
                        memcpy(vflcxtblock, &(*cxt).vflcxtblocks, 8);
                        minusn = 0xFFFFFFFF;
                        vflcxtidx = 4;
                        for (k = 0; k < 4; k++)
                            if (vflcxtblock[k] != 0xFFFF)
                                if (ftl_vfl_read_page(i, vflcxtblock[k], 0,
                                                      ftl_buffer,
                                                      &ftl_sparebuffer) == 0)
                                    if (ftl_sparebuffer.meta.usn > 0
                                     && ftl_sparebuffer.meta.usn <= minusn)
                                    {
                                        minusn = ftl_sparebuffer.meta.usn;
                                        vflcxtidx = k;
                                    }
                        if (vflcxtidx == 4)
                        {
                            debug_printf("VFL: No CXT block found on bank %u!\n", i);
                            return 1;
                        }
                        debug_printf("VFL: CXT block found at page %u on bank %u, "
                                     "searching for most recent CXT...\n",
                                     vflcxtblock[vflcxtidx], i);
                        last = 0;
                        uint32_t max = (*ftl_nand_type).pagesperblock;
                        for (k = 8; k < max; k += 8)
                        {
                            if (ftl_vfl_read_page(i, vflcxtblock[vflcxtidx],
                                                  k, ftl_buffer,
                                                  &ftl_sparebuffer) != 0)
                                break;
                            last = k;
                        }
                        if (ftl_vfl_read_page(i, vflcxtblock[vflcxtidx],
                                              last, ftl_buffer,
                                              &ftl_sparebuffer) != 0)
                        {
                            debug_printf("VFL: Re-reading VFL CXT block "
                                         "on bank %u failed!?\n", (unsigned)i);
                            return 1;
                        }
                        memcpy(&ftl_vfl_cxt[i], ftl_buffer, 0x800);
                        if (ftl_vfl_verify_checksum(i) != 0) return 1;
                        break;
                    }
        }
        else
        {
            debug_printf("VFL: Couldn't load bank %u lowlevel BBT!\n", (unsigned)i);
            return 1;
        }
    }
    cxt = ftl_vfl_get_newest_cxt();
    debug_printf("VFL: Initialization completed.\n");
    debug_printf("VFL: Pointing to FTL CXT blocks: %u, %u, %u\n",
                 (*cxt).ftlctrlblocks[0], (*cxt).ftlctrlblocks[1], (*cxt).ftlctrlblocks[2]);
    return 0;
}


/* Mounts the actual FTL */
uint32_t ftl_open(void)
{
    uint32_t i;
    uint32_t ret;
    uint32_t ppb = (*ftl_nand_type).pagesperblock * ftl_banks;
    struct ftl_vfl_cxt_type* cxt = ftl_vfl_get_newest_cxt();

    uint32_t ftlcxtblock = 0xffffffff;
    uint32_t minusn = 0xffffffff;
    for (i = 0; i < 3; i++)
    {
        debug_printf("FTL: Searching for CXT at block %u...\n", (*cxt).ftlctrlblocks[i]);
        ret = ftl_vfl_read(ppb * (*cxt).ftlctrlblocks[i],
                           ftl_buffer, &ftl_sparebuffer, 1, 0);
        if ((ret &= 0x11F) != 0)
        {
            debug_printf("FTL: Read error, trying next CXT block...\n");
            continue;
        }
        if (ftl_sparebuffer.user.type - 0x43 > 4)
        {
            debug_printf("FTL: Hit a non-meta page (type 0x%2X), trying next CXT...\n",
                         ftl_sparebuffer.user.type);
            continue;
        }
        if (ftlcxtblock != 0xffffffff && ftl_sparebuffer.meta.usn >= minusn)
        {
            debug_printf("FTL: Ignoring valid CXT block (USN %8X), "
                         "we already got a better one\n", ftl_sparebuffer.meta.usn);
            continue;
        }
        debug_printf("FTL: Found valid CXT block (USN %8X)\n", ftl_sparebuffer.meta.usn);
        minusn = ftl_sparebuffer.meta.usn;
        ftlcxtblock = (*cxt).ftlctrlblocks[i];
    }

    if (ftlcxtblock == 0xffffffff)
    {
        debug_printf("FTL: Cannot find FTL context block!\n");
        return 1;
    }

    debug_printf("FTL: Found FTL context block: vBlock %d\n", ftlcxtblock);

    debug_printf("FTL: Scanning pages of FTL CXT block...\n");
    uint32_t ftlcxtfound = 0;
    for (i = (*ftl_nand_type).pagesperblock * ftl_banks - 1; i > 0; i--)
    {
        ret = ftl_vfl_read(ppb * ftlcxtblock + i,
                           ftl_buffer, &ftl_sparebuffer, 1, 0);
        if ((ret & 0x11F) != 0) continue;
        else if (ftl_sparebuffer.user.type == 0x43)
        {
            debug_printf("FTL: Found most recent CXT at page %u\n", i);
            memcpy(&ftl_cxt, ftl_buffer, 0x28C);
            ftlcxtfound = 1;
            break;
        }
        else
        {
            debug_printf("FTL: Found unclean shutdown mark!\n");
            debug_printf("FTL: Forcing mount nevertheless...\n");
        }
    }

    if (ftlcxtfound == 0)
    {
        debug_printf("FTL: Couldn't find FTL CXT page!\n");
        return 1;
    }

    debug_printf("FTL: Successfully read FTL context\n");

    uint32_t pagestoread = (*ftl_nand_type).userblocks >> 10;
    if (((*ftl_nand_type).userblocks & 0x1FF) != 0) pagestoread++;

    debug_printf("FTL: Reading block map...\n");
    for (i = 0; i < pagestoread; i++)
    {
        if ((ftl_vfl_read(ftl_cxt.ftl_map_pages[i],
                          ftl_buffer, &ftl_sparebuffer, 1, 1) & 0x11F) != 0)
        {
            debug_printf("FTL: Failed to read block map page %u!\n", (unsigned)i);
            return 1;
        }

        uint32_t toread = 2048;
        if (toread > ((*ftl_nand_type).userblocks << 1) - (i << 11))
            toread = ((*ftl_nand_type).userblocks << 1) - (i << 11);

        memcpy(&ftl_map[i << 10], ftl_buffer, toread);
    }

    pagestoread = ((*ftl_nand_type).userblocks + 23) >> 10;
    if ((((*ftl_nand_type).userblocks + 23) & 0x1FF) != 0) pagestoread++;

    debug_printf("FTL: Reading erase counters...\n");
    for (i = 0; i < pagestoread; i++)
    {
        if ((ftl_vfl_read(ftl_cxt.ftl_erasectr_pages[i],
                          ftl_buffer, &ftl_sparebuffer, 1, 1) & 0x11F) != 0)
        {
            debug_printf("FTL: Failed to read erase counter page %u!\n", (unsigned)i);
            return 1;
        }

        uint32_t toread = 2048;
        if (toread > (((*ftl_nand_type).userblocks + 23) << 1) - (i << 11))
            toread = (((*ftl_nand_type).userblocks + 23) << 1) - (i << 11);

        memcpy(&ftl_erasectr[i << 10], ftl_buffer, toread);
    }

    return 0;
}


/* Exposed function: Read highlevel sectors */
uint32_t ftl_read(uint32_t sector, uint32_t count, void* buffer)
{
    uint32_t i;
    uint32_t ppb = (*ftl_nand_type).pagesperblock * ftl_banks;
    uint32_t error = 0;

#ifdef FTL_TRACE
    debug_printf("FTL: Reading %d sectors starting at %d\n", count, sector);
#endif

    if (sector + count > (*ftl_nand_type).userblocks * ppb)
    {
        debug_printf("FTL: Sector %d is out of range!\n", sector + count - 1);
        return 1;
    }
    if (count == 0) return 0;

    for (i = 0; i < count; i++)
    {
        uint32_t block = (sector + i) / ppb;
        uint32_t page = (sector + i) % ppb;

        uint32_t abspage = ftl_map[block] * ppb + page;

        uint32_t ret = ftl_vfl_read(abspage, &((uint8_t*)buffer)[i << 11],
                                    &ftl_sparebuffer, 1, 1);
        if ((ret & 2) != 0) memset(&((uint8_t*)buffer)[i << 11], 0, 0x800);
        else if ((ret & 0x11D) != 0 || ftl_sparebuffer.user.eccmark != 0xFF)
        {
            debug_printf("FTL: Error while reading sector %d!\n", (sector + i));
            error = 1;
            memset(&((uint8_t*)buffer)[i << 11], 0, 0x800);
        }
    }
    return error;
}


/* Initializes and mounts the FTL.
   As long as nothing was written, you won't need to unmount it.
   Before shutting down after writing something, call ftl_sync(),
   which will just do nothing if everything was already clean. */
uint32_t ftl_init(void)
{
    uint32_t i;
    uint32_t result = 0;
    uint32_t foundsignature, founddevinfo, blockwiped, repaired, skip;
    if (nand_init() != 0)
    {
        debug_printf("FTL: Lowlevel NAND driver init failed!\n");
        return 1;
    }
    ftl_banks = 0;
    for (i = 0; i < 4; i++)
        if (nand_get_device_type(i) != 0) ftl_banks = i + 1;
    ftl_nand_type = nand_get_device_type(0);
    foundsignature = 0;
    blockwiped = 1;
    debug_printf("FTL: Searching for signature...\n");
    for (i = 0; i < (*ftl_nand_type).pagesperblock; i++)
    {
        result = nand_read_page(0, i, ftl_buffer, (uint32_t*)0, 1, 1);
        if ((result & 0x11F) == 0)
        {
            blockwiped = 0;
            if (((uint32_t*)ftl_buffer)[0] != 0x41303034) continue;
            debug_printf("FTL: Found Whimory V4 header signature\n");
            foundsignature = 1;
            break;
        }
        else if ((result & 2) != 2) blockwiped = 0;
    }
    debug_printf("FTL: Searching for DEVICEINFO...\n");
    founddevinfo = ftl_has_devinfo();

    repaired = 0;
    skip = 0;
    if (founddevinfo == 0)
    {
        debug_printf("FTL: No DEVICEINFO found on at least one bank!\n");
        return 1;
    }
    if (foundsignature != 0 && (result & 0x11F) != 0)
    {
        debug_printf("FTL: Problem with the signature!\n");
        return 1;
    }

    if (ftl_vfl_open() == 0)
        if (ftl_open() == 0) return 0;

    debug_printf("FTL: Initialization failed!\n");
    return 1;
}


void main(void)
{
    uint32_t i, j, k;

    debug_printf("iPod Nano 2G FTL consistency checker v0.1.0\n");
    debug_printf("Copyright 2010 TheSeven\n\n");
    debug_printf("Trying to mount the FTL...\n");
    if (ftl_init())
    {
        debug_printf("Something went wrong while mounting the FTL.\n");
        debug_printf("See messages above for details.\n");
    }
    else
    {
        debug_printf("Successfully mounted the FTL.\n\n");

        for (i = 0; i < ftl_banks; i++)
        {
            debug_printf("LL BBT dump (bank %u):\n", i);
            hexdump(&ftl_bbt[i], 0x410);
            debug_printf("\n");
        }

        for (i = 0; i < ftl_banks; i++)
        {
            debug_printf("VFL context dump (bank %u):\n", i);
            hexdump(&ftl_vfl_cxt[i], sizeof(struct ftl_vfl_cxt_type));
            debug_printf("\n");
        }

        debug_printf("FTL context dump:\n");
        hexdump(&ftl_cxt, sizeof(struct ftl_cxt_type));
        debug_printf("\n");

        debug_printf("Block map dump:\n");
        hexdump(&ftl_map, 0x4000);
        debug_printf("\n");

        debug_printf("Erase counter dump:\n");
        hexdump(&ftl_erasectr, 0x4000);
        debug_printf("\n");

        for (i = 0; i < ftl_banks; i++)
        {
            uint32_t badblocks = 0;
            for (j = 0; j < (*ftl_nand_type).blocks >> 3; j++)
            {
                uint8_t bbtentry = ftl_bbt[i][j];
                for (k = 0; k < 8; k++) if ((bbtentry & (1 << k)) == 0) badblocks++;
            }
            debug_printf("LL BBT for bank %d: %d bad blocks\n", i, badblocks);
            badblocks = 0;
            for (j = 0; j < ftl_vfl_cxt[i].sparecount; j++)
                if (ftl_vfl_cxt[i].remaptable[j] == 0xFFFF) badblocks++;
            debug_printf("Bank %d: %d of %d spare blocks are bad\n",
                         i, badblocks, ftl_vfl_cxt[i].sparecount);
            debug_printf("Bank %d: %d blocks remapped\n",
                         i, ftl_vfl_cxt[i].spareused);
            debug_printf("Bank %d: %d blocks scheduled for remapping\n",
                         i, 0x334 - ftl_vfl_cxt[i].scheduledstart);
        }
        uint32_t min = 0xFFFFFFFF, max = 0, total = 0;
        for (i = 0; i < (*ftl_nand_type).userblocks + 23; i++)
        {
            if (ftl_erasectr[i] > max) max = ftl_erasectr[i];
            if (ftl_erasectr[i] < min) min = ftl_erasectr[i];
            total += ftl_erasectr[i];
        }
        debug_printf("Erase counters: Minimum: %d, maximum %d, average: %d, total: %d\n",
                     min, max, total / ((*ftl_nand_type).userblocks + 23), total);

        debug_printf("\n");
        debug_printf("LL BBT bad blocks:\n");
        for (i = 0; i < ftl_banks; i++)
        {
            debug_printf("Bank %u:", i);
            for (j = 0; j < (*ftl_nand_type).blocks >> 3; j++)
            {
                uint8_t bbtentry = ftl_bbt[i][j];
                for (k = 0; k < 8; k++)
                    if ((bbtentry & (1 << k)) == 0)
                        debug_printf(" %u", (j << 3) | k);
            }
            debug_printf("\n");
        }

        debug_printf("\n");
        debug_printf("VFL remapped blocks:\n");
        for (i = 0; i < ftl_banks; i++)
        {
            debug_printf("Bank %u:", i);
            for (j = 0; j < ftl_vfl_cxt[i].sparecount; j++)
                if (ftl_vfl_cxt[i].remaptable[j] == 0xFFFF)
                    debug_printf(" BAD:%u", j);
                else if (ftl_vfl_cxt[i].remaptable[j])
                    debug_printf(" %u=>%u", ftl_vfl_cxt[i].remaptable[j], j);
            for (j = ftl_vfl_cxt[i].scheduledstart; j < 0x334; j++)
                debug_printf(" SCHEDULED:%u", ftl_vfl_cxt[i].remaptable[j]);
            debug_printf("\n");
        }

        debug_printf("\n");
        debug_printf("Block pool:");
        for (i = 0; i < 0x14; i++)
            debug_printf(" %u", ftl_cxt.blockpool[i]);
        debug_printf("\n");
        debug_printf("Block pool free count: %u\n", ftl_cxt.freecount);
        debug_printf("Block pool ring buffer index: %u\n", ftl_cxt.nextfreeidx);
        debug_printf("Control block pool:");
        for (i = 0; i < 3; i++)
            debug_printf(" %u", ftl_cxt.ftlctrlblocks[i]);
        debug_printf("\n");
        struct ftl_vfl_cxt_type* cxt = ftl_vfl_get_newest_cxt();
        debug_printf("Control block pool (from VFL):");
        for (i = 0; i < 3; i++)
            debug_printf(" %u", (*cxt).ftlctrlblocks[i]);
        debug_printf("\n");
        for (i = 0; i < 3; i++)
            if (ftl_cxt.ftlctrlblocks[i] != (*cxt).ftlctrlblocks[i])
                debug_printf("VFL points to a wrong FTL CXT block! (%u vs. %u)\n",
                             ftl_cxt.ftlctrlblocks[i], (*cxt).ftlctrlblocks[i]);

        uint32_t alloc[0x100];
        for (i = 0; i < 0x100; i++) alloc[i] = 0;
        for (i = 0; i < (*ftl_nand_type).userblocks; i++)
            if (alloc[ftl_map[i] >> 5] & (1 << (ftl_map[i] & 0x1f)))
                debug_printf("Block allocated multiple times: %u => %u!\n", i, ftl_map[i]);
            else alloc[ftl_map[i] >> 5] |= 1 << (ftl_map[i] & 0x1f);
        for (i = 0; i < 0x14; i++)
            if (alloc[ftl_cxt.blockpool[i] >> 5] & (1 << (ftl_cxt.blockpool[i] & 0x1f)))
                debug_printf("Block allocated multiple times: POOL => %u!\n", ftl_cxt.blockpool[i]);
            else alloc[ftl_cxt.blockpool[i] >> 5] |= 1 << (ftl_cxt.blockpool[i] & 0x1f);
        for (i = 0; i < 3; i++)
            if (alloc[ftl_cxt.ftlctrlblocks[i] >> 5] & (1 << (ftl_cxt.ftlctrlblocks[i] & 0x1f)))
                debug_printf("Block allocated multiple times: CTRL => %u!\n", ftl_cxt.ftlctrlblocks[i]);
            else alloc[ftl_cxt.ftlctrlblocks[i] >> 5] |= 1 << (ftl_cxt.ftlctrlblocks[i] & 0x1f);
        for (i = 0; i < (*ftl_nand_type).userblocks + 0x17; i++)
            if (!(alloc[i >> 5] & (1 << (i & 0x1f))))
                debug_printf("Dangling block: %u!\n", i);

        debug_printf("\n");
        debug_printf("done\n");
    }
}


void irqhandler(void) __attribute__((interrupt("IRQ")));
void irqhandler(void)
{
    uint32_t sources;
    sources = SRCPND;

    SRCPND = sources;
    INTPND = sources;
}
