//
//
//    Copyright 2009-2010 TheSeven
//
//
//    This file is part of the Linux4Nano toolkit.
//
//    TheSeven's iBugger is free software: you can redistribute it and/or
//    modify it under the terms of the GNU General Public License as
//    published by the Free Software Foundation, either version 2 of the
//    License, or (at your option) any later version.
//
//    TheSeven's iBugger is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//    See the GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with the Linux4Nano toolkit.  If not, see <http://www.gnu.org/licenses/>.
//
//



#include <toolkit.h>
#include <nand.h>
#include <ftl.h>



/* Keeps the state of the FTL, both on flash and in memory */
struct ftl_cxt_type
{

    /* Update sequence number of the FTL context, decremented
       every time a new revision of FTL meta data is written. */
    uint32_t usn;

    /* Update sequence number for user data blocks. Incremented
       every time a portion of user pages is written, so that
       a consistency check can determine which copy of a user
       page is the most recent one. */
    uint32_t nextblockusn;

    /* Count of currently free pages in the block pool */
    uint16_t freecount;

    /* Index to the first free hyperblock in the blockpool ring buffer */
    uint16_t nextfreeidx;

    /* This is a counter that is used to better distribute block
       wear. It is incremented on every block erase, and if it
       gets too high (300 on writes, 20 on sync), the most and
       least worn hyperblock will be swapped (causing an additional
       block write) and the counter will be decreased by 20. */
    uint16_t swapcounter;

    /* Ring buffer of currently free hyperblocks. nextfreeidx is the
       index to freecount free ones, the other ones are currently
       allocated for scattered page hyperblocks. */
    uint16_t blockpool[0x14];

    /* Alignment to 32 bits */
    uint16_t field_36;

    /* vPages where the block map is stored */
    uint32_t ftl_map_pages[8];

    /* Probably additional map page number space for bigger chips */
    uint8_t field_58[0x28];

    /* vPages where the erase counters are stored */
    uint32_t ftl_erasectr_pages[8];

    /* Seems to be padding */
    uint8_t field_A0[0x70];

    /* Pointer to ftl_map used by Whimory, not used by us */
    uint32_t ftl_map_ptr;

    /* Pointer to ftl_erasectr used by Whimory, not used by us */
    uint32_t ftl_erasectr_ptr;

    /* Pointer to ftl_log used by Whimory, not used by us */
    uint32_t ftl_log_ptr;

    /* Flag used to indicate that some erase counter pages should be committed
       because they were changed more than 100 times since the last commit. */
    uint32_t erasedirty;

    /* Seems to be unused */
    uint16_t field_120;

    /* vBlocks used to store the FTL context, map, and erase
       counter pages. This is also a ring buffer, and the oldest
       page gets swapped with the least used page from the block
       pool ring buffer when a new one is allocated. */
    uint16_t ftlctrlblocks[3];

    /* The last used vPage number from ftlctrlblocks */
    uint32_t ftlctrlpage;

    /* Set on context sync, reset on write, so obviously never
       zero in the context written to the flash */
    uint32_t clean_flag;

    /* Seems to be unused, but gets loaded from flash by Whimory. */
    uint8_t field_130[0x6d0];

} __attribute__((packed));


/* Keeps the state of the bank's VFL, both on flash and in memory.
   There is one of these per bank. */
struct ftl_vfl_cxt_type
{

    /* Cross-bank update sequence number, incremented on every VFL
       context commit on any bank. */
    uint32_t usn;

    /* See ftl_cxt.ftlctrlblocks. This is stored to the VFL contexts
       in order to be able to find the most recent FTL context copy
       when mounting the FTL. The VFL context number this will be
       written to on an FTL context commit is chosen semi-randomly. */
    uint16_t ftlctrlblocks[3];

    /* Alignment to 32 bits */
    uint8_t field_A[2];

    /* Decrementing update counter for VFL context commits per bank */
    uint32_t updatecount;

    /* Number of the currently active VFL context block, it's an index
       into vflcxtblocks. */
    uint16_t activecxtblock;

    /* Number of the first free page in the active VFL context block */
    uint16_t nextcxtpage;

    /* Seems to be unused */
    uint8_t field_14[4];

    /* Incremented every time a block erase error leads to a remap,
       but doesn't seem to be read anywhere. */
    uint16_t field_18;

    /* Number of spare blocks used */
    uint16_t spareused;

    /* pBlock number of the first spare block */
    uint16_t firstspare;

    /* Total number of spare blocks */
    uint16_t sparecount;

    /* Block remap table. Contains the vBlock number the n-th spare
       block is used as a replacement for. 0 = unused, 0xFFFF = bad. */
    uint16_t remaptable[0x334];

    /* Bad block table. Each bit represents 8 blocks. 1 = OK, 0 = Bad.
       If the entry is zero, you should look at the remap table to see
       if the block is remapped, and if yes, where the replacement is. */
    uint8_t bbt[0x11A];

    /* pBlock numbers used to store the VFL context. This is a ring
       buffer. On a VFL context write, always 8 pages are written,
       and it passes if at least 4 of them can be read back. */
    uint16_t vflcxtblocks[4];

    /* Blocks scheduled for remapping are stored at the end of the
       remap table. This is the first index used for them. */
    uint16_t scheduledstart;

    /* Probably padding */
    uint8_t field_7AC[0x4C];

    /* First checksum (addition) */
    uint32_t checksum1;

    /* Second checksum (XOR), there is a bug in whimory regarding this. */
    uint32_t checksum2;

} __attribute__((packed));


/* Layout of the spare bytes of each page on the flash */
union ftl_spare_data_type
{

    /* The layout used for actual user data (types 0x40 and 0x41) */
    struct ftl_spare_data_user_type
    {

        /* The lPage, i.e. Sector, number */
        uint32_t lpn;

        /* The update sequence number of that page,
           copied from ftl_cxt.nextblockusn on write */
        uint32_t usn;

        /* Seems to be unused */
        uint8_t field_8;

        /* Type field, 0x40 (data page) or 0x41
           (last data page of hyperblock) */
        uint8_t type;

        /* ECC mark, usually 0xFF. If an error occurred while reading the
           page during a copying operation earlier, this will be 0x55. */
        uint8_t eccmark;

        /* Seems to be unused */
        uint8_t field_B;

        /* ECC data for the user data */
        uint8_t dataecc[0x28];

        /* ECC data for the first 0xC bytes above */
        uint8_t spareecc[0xC];

    } __attribute__((packed)) user;

    /* The layout used for meta data (other types) */
    struct ftl_spare_data_meta_type
    {

        /* ftl_cxt.usn for FTL stuff, ftl_vfl_cxt.updatecount for VFL stuff */
        uint32_t usn;

        /* Index of the thing inside the page,
           for example number / index of the map or erase counter page */
        uint16_t idx;

        /* Seems to be unused */
        uint8_t field_6;

        /* Seems to be unused */
        uint8_t field_7;

        /* Seems to be unused */
        uint8_t field_8;

       /* Type field:
            0x43: FTL context page
            0x44: Block map page
            0x46: Erase counter page
            0x47: "FTL is currently mounted", i.e. unclean shutdown, mark
            0x80: VFL context page */
        uint8_t type;

        /* ECC mark, usually 0xFF. If an error occurred while reading the
           page during a copying operation earlier, this will be 0x55. */
        uint8_t eccmark;

        /* Seems to be unused */
        uint8_t field_B;

        /* ECC data for the user data */
        uint8_t dataecc[0x28];

        /* ECC data for the first 0xC bytes above */
        uint8_t spareecc[0xC];

    } __attribute__((packed)) meta;

};


/* Pointer to an info structure regarding the flash type used */
const struct nand_device_info_type* ftl_nand_type;

/* Number of banks we detected a chip on */
uint32_t ftl_banks;

/* Block map, used vor pBlock to vBlock mapping */
uint16_t ftl_map[0x2000] __attribute__((aligned(16)));
uint32_t blk_usn[0x2000];
uint8_t blk_type[0x2000];
uint32_t erasectr_usn[8];
uint32_t pageusn[0x200];
uint8_t pagedata[0x200][0x800];

/* VFL context for each bank */
struct ftl_vfl_cxt_type ftl_vfl_cxt[4];

/* FTL context */
struct ftl_cxt_type ftl_cxt __attribute__((aligned(16)));

/* Temporary data buffers for internal use by the FTL */
uint8_t ftl_buffer[0x800] __attribute__((aligned(16)));

/* Temporary spare byte buffer for internal use by the FTL */
union ftl_spare_data_type ftl_sparebuffer __attribute__((aligned(16)));
union ftl_spare_data_type ftl_sparebuffer2 __attribute__((aligned(16)));

/* Global cross-bank update sequence number of the VFL context */
uint32_t ftl_vfl_usn;
uint32_t meta_usn;
uint32_t user_usn;

uint32_t allocmode;
uint32_t firstfree;

/* Erase counters for the vBlocks */
uint16_t ftl_erasectr[0x2000] __attribute__((aligned(16)));

uint16_t lcdbuffer[0x5AC0];
uint32_t statusbar_y;
uint32_t statusbar_offset;
uint32_t statusbar_divisor;
uint32_t statusbar_multiplier;
uint32_t statusbar_current;



/* Finds a device info page for the specified bank and returns its number.
   Used to check if one is present, and to read the lowlevel BBT. */
uint32_t ftl_find_devinfo(uint32_t bank)
{
    /* Scan the last 10% of the flash for device info pages */
    uint32_t lowestBlock = (*ftl_nand_type).blocks
                         - ((*ftl_nand_type).blocks / 10);
    uint32_t block, page, pagenum;
    for (block = (*ftl_nand_type).blocks - 1; block >= lowestBlock; block--)
    {
        page = (*ftl_nand_type).pagesperblock - 8;
        for (; page < (*ftl_nand_type).pagesperblock; page++)
        {
            pagenum = block * (*ftl_nand_type).pagesperblock + page;
            if ((nand_read_page(bank, pagenum, ftl_buffer,
                                &ftl_sparebuffer, 1, 0) & 0x11F) != 0)
                continue;
            if (memcmp(ftl_buffer, "DEVICEINFOSIGN\0", 0x10) == 0)
                return pagenum;
        }
    }
    return 0;
}


/* Checks if all banks have proper device info pages */
uint32_t ftl_has_devinfo(void)
{
    uint32_t i;
    for (i = 0; i < ftl_banks; i++) if (ftl_find_devinfo(i) == 0) return 0;
    return 1;
}


/* Loads the lowlevel BBT for a bank to the specified buffer.
   This is based on some cryptic disassembly and not fully understood yet. */
uint32_t ftl_load_bbt(uint32_t bank, uint8_t* bbt)
{
    uint32_t i, j;
    uint32_t pagebase, page = ftl_find_devinfo(bank), page2;
    uint32_t unk1, unk2, unk3;
    if (page == 0) return 1;
    pagebase = page & ~((*ftl_nand_type).pagesperblock - 1);
    if (nand_read_page(bank, page, ftl_buffer,
                       (uint32_t*)0, 1, 0) & 0x11F) return 1;
    if (memcmp(&ftl_buffer[0x18], "BBT", 4)) return 1;
    unk1 = ((uint16_t*)ftl_buffer)[0x10];
    unk2 = ((uint16_t*)ftl_buffer)[0x11];
    unk3 = ((uint16_t*)ftl_buffer)[((uint32_t*)ftl_buffer)[4] * 6 + 10]
         + ((uint16_t*)ftl_buffer)[((uint32_t*)ftl_buffer)[4] * 6 + 11];
    for (i = 0; i < unk1; i++)
    {
        for (j = 0; ; j++)
        {
            page2 = unk2 + i + unk3 * j;
            if (page2 >= (uint32_t)((*ftl_nand_type).pagesperblock - 8))
                break;
            if ((nand_read_page(bank, pagebase + page2, ftl_buffer,
                                (void*)0, 1, 0) & 0x11F) == 0)
            {
                memcpy(bbt, ftl_buffer, 0x410);
                return 0;
            }
        }
    }
    return 1;
}


/* Calculates the checksums for the VFL context page of the specified bank */
void ftl_vfl_calculate_checksum(uint32_t bank,
                                uint32_t* checksum1, uint32_t* checksum2)
{
    uint32_t i;
    *checksum1 = 0xAABBCCDD;
    *checksum2 = 0xAABBCCDD;
    for (i = 0; i < 0x1FE; i++)
    {
        *checksum1 += ((uint32_t*)(&ftl_vfl_cxt[bank]))[i];
        *checksum2 ^= ((uint32_t*)(&ftl_vfl_cxt[bank]))[i];
    }
}


/* Checks if the checksums of the VFL context
   of the specified bank are correct */
uint32_t ftl_vfl_verify_checksum(uint32_t bank)
{
    uint32_t checksum1, checksum2;
    ftl_vfl_calculate_checksum(bank, &checksum1, &checksum2);
    if (checksum1 == ftl_vfl_cxt[bank].checksum1) return 0;
    /* The following line is pretty obviously a bug in Whimory,
       but we do it the same way for compatibility. */
    if (checksum2 != ftl_vfl_cxt[bank].checksum2) return 0;
    return 1;
}


/* Updates the checksums of the VFL context of the specified bank */
void ftl_vfl_update_checksum(uint32_t bank)
{
    ftl_vfl_calculate_checksum(bank, &ftl_vfl_cxt[bank].checksum1,
                               &ftl_vfl_cxt[bank].checksum2);
}


/* Writes 8 copies of the VFL context of the specified bank to flash,
   and succeeds if at least 4 can be read back properly. */
uint32_t ftl_vfl_store_cxt(uint32_t bank)
{
    uint32_t i;
    ftl_vfl_cxt[bank].updatecount--;
    ftl_vfl_cxt[bank].usn = ++ftl_vfl_usn;
    ftl_vfl_cxt[bank].nextcxtpage += 8;
    ftl_vfl_update_checksum(bank);
    memset(&ftl_sparebuffer, 0xFF, 0x40);
    ftl_sparebuffer.meta.usn = ftl_vfl_cxt[bank].updatecount;
    ftl_sparebuffer.meta.field_8 = 0;
    ftl_sparebuffer.meta.type = 0x80;
    for (i = 1; i <= 8; i++)
    {
        uint32_t index = ftl_vfl_cxt[bank].activecxtblock;
        uint32_t block = ftl_vfl_cxt[bank].vflcxtblocks[index];
        uint32_t page = block * (*ftl_nand_type).pagesperblock;
        page += ftl_vfl_cxt[bank].nextcxtpage - i;
        nand_write_page(bank, page, &ftl_vfl_cxt[bank], &ftl_sparebuffer, 1);
    }
    uint32_t good = 0;
    for (i = 0; i < 8; i++)
    {
        uint32_t index = ftl_vfl_cxt[bank].activecxtblock;
        uint32_t block = ftl_vfl_cxt[bank].vflcxtblocks[index];
        uint32_t page = block * (*ftl_nand_type).pagesperblock;
        page += ftl_vfl_cxt[bank].nextcxtpage - i;
        if ((nand_read_page(bank, page, ftl_buffer,
                            &ftl_sparebuffer, 1, 0) & 0x11F) != 0)
            continue;
        if (memcmp(ftl_buffer, &ftl_vfl_cxt[bank], 0x7AC) != 0)
            continue;
        if (ftl_sparebuffer.meta.usn != ftl_vfl_cxt[bank].updatecount)
            continue;
        if (ftl_sparebuffer.meta.field_8 == 0
         && ftl_sparebuffer.meta.type == 0x80) good++;
    }
    return good > 3 ? 0 : 1;
}


/* Commits the VFL context of the specified bank to flash,
   retries until it works or all available pages have been tried */
uint32_t ftl_vfl_commit_cxt(uint32_t bank)
{
    if (ftl_vfl_cxt[bank].nextcxtpage + 8 <= (*ftl_nand_type).pagesperblock)
        if (ftl_vfl_store_cxt(bank) == 0) return 0;
    uint32_t current = ftl_vfl_cxt[bank].activecxtblock;
    uint32_t i = current, j;
    while (1)
    {
        i = (i + 1) & 3;
        if (i == current) break;
        if (ftl_vfl_cxt[bank].vflcxtblocks[i] == 0xFFFF) continue;
        for (j = 0; j < 4; j++)
            if (nand_block_erase(bank, ftl_vfl_cxt[bank].vflcxtblocks[i]
                                     * (*ftl_nand_type).pagesperblock) == 0)
                break;
        if (j == 4) continue;
        ftl_vfl_cxt[bank].activecxtblock = i;
        ftl_vfl_cxt[bank].nextcxtpage = 0;
        if (ftl_vfl_store_cxt(bank) == 0) return 0;
    }
    return 1;
}


/* Returns a pointer to the most recently updated VFL context,
   used to find out the current FTL context vBlock numbers
   (planetbeing's "maxthing") */
struct ftl_vfl_cxt_type* ftl_vfl_get_newest_cxt(void)
{
    uint32_t i, maxusn;
    struct ftl_vfl_cxt_type* cxt = (struct ftl_vfl_cxt_type*)0;
    maxusn = 0;
    for (i = 0; i < ftl_banks; i++)
        if (ftl_vfl_cxt[i].usn >= maxusn)
        {
            cxt = &ftl_vfl_cxt[i];
            maxusn = ftl_vfl_cxt[i].usn;
        }
    return cxt;
}


/* Checks if the specified pBlock is marked bad in the supplied lowlevel BBT.
   Only used while mounting the VFL. */
uint32_t ftl_is_good_block(uint8_t* bbt, uint32_t block)
{
    if ((bbt[block >> 3] & (1 << (block & 7))) == 0) return 0;
    else return 1;
}


/* Checks if the specified vBlock could be remapped */
uint32_t ftl_vfl_is_good_block(uint32_t bank, uint32_t block)
{
    uint8_t bbtentry = ftl_vfl_cxt[bank].bbt[block >> 6];
    if ((bbtentry & (1 << ((7 - (block >> 3)) & 7))) == 0) return 0;
    else return 1;
}


/* Tries to read a VFL context from the specified bank, pBlock and page */
uint32_t ftl_vfl_read_page(uint32_t bank, uint32_t block,
                           uint32_t startpage, void* databuffer,
                           union ftl_spare_data_type* sparebuffer)
{
    uint32_t i;
    for (i = 0; i < 8; i++)
    {
        uint32_t page = block * (*ftl_nand_type).pagesperblock
                      + startpage + i;
        if ((nand_read_page(bank, page, databuffer,
                            sparebuffer, 1, 1) & 0x11F) == 0)
            if ((*sparebuffer).meta.field_8 == 0
             && (*sparebuffer).meta.type == 0x80)
                return 0;
    }
    return 1;
}


/* Translates a bank and vBlock to a pBlock, following remaps */
uint32_t ftl_vfl_get_physical_block(uint32_t bank, uint32_t block)
{
    if (ftl_vfl_is_good_block(bank, block) == 1) return block;

    uint32_t spareindex;
    uint32_t spareused = ftl_vfl_cxt[bank].spareused;
    for (spareindex = 0; spareindex < spareused; spareindex++)
        if (ftl_vfl_cxt[bank].remaptable[spareindex] == block)
            return ftl_vfl_cxt[bank].firstspare + spareindex;
    return block;
}


// Reads the specified vPage, dealing with all kinds of trouble
uint32_t ftl_vfl_read(uint32_t vpage, void* buffer, void* sparebuffer,
                      uint32_t checkempty, uint32_t remaponfail)
{
    uint32_t ppb = (*ftl_nand_type).pagesperblock * ftl_banks;
    uint32_t syshyperblocks = (*ftl_nand_type).blocks
                            - (*ftl_nand_type).userblocks - 0x17;
    uint32_t abspage = vpage + ppb * syshyperblocks;
    if (abspage >= (*ftl_nand_type).blocks * ppb || abspage < ppb)
        return 4;

    uint32_t bank = abspage % ftl_banks;
    uint32_t block = abspage / ppb;
    uint32_t page = (abspage / ftl_banks) % (*ftl_nand_type).pagesperblock;
    uint32_t physblock = ftl_vfl_get_physical_block(bank, block);
    uint32_t physpage = physblock * (*ftl_nand_type).pagesperblock + page;

    uint32_t ret = nand_read_page(bank, physpage, buffer,
                                  sparebuffer, 1, checkempty);

    if ((ret & 0x11D) != 0 && (ret & 2) == 0)
    {
        nand_reset(bank);
        ret = nand_read_page(bank, physpage, buffer,
                             sparebuffer, 1, checkempty);
        return ret;
    }

    return ret;
}


/* Writes the specified vPage, dealing with all kinds of trouble */
uint32_t ftl_vfl_write(uint32_t vpage, void* buffer, void* sparebuffer)
{
    uint32_t ppb = (*ftl_nand_type).pagesperblock * ftl_banks;
    uint32_t syshyperblocks = (*ftl_nand_type).blocks
                            - (*ftl_nand_type).userblocks - 0x17;
    uint32_t abspage = vpage + ppb * syshyperblocks;
    if (abspage >= (*ftl_nand_type).blocks * ppb || abspage < ppb)
        return 4;

    uint32_t bank = abspage % ftl_banks;
    uint32_t block = abspage / ppb;
    uint32_t page = (abspage / ftl_banks) % (*ftl_nand_type).pagesperblock;
    uint32_t physblock = ftl_vfl_get_physical_block(bank, block);
    uint32_t physpage = physblock * (*ftl_nand_type).pagesperblock + page;

    if (nand_write_page(bank, physpage, buffer, sparebuffer, 1) == 0)
        return 0;
    if ((nand_read_page(bank, physpage, ftl_buffer,
                        &ftl_sparebuffer, 1, 1) & 0x11F) == 0)
        return 0;
    return 1;
}


/* Mounts the VFL on all banks */
uint32_t ftl_vfl_open(void)
{
    uint32_t i, j, k;
    uint32_t minusn, vflcxtidx, last;
    uint16_t vflcxtblock[4];

    uint32_t syshyperblocks = (*ftl_nand_type).blocks
                            - (*ftl_nand_type).userblocks - 0x18;
    uint8_t bbt[0x410];
    ftl_vfl_usn = 0;

    for (i = 0; i < ftl_banks; i++)
    {
        if (!ftl_load_bbt(i, bbt))
        {
            for (j = 1; j <= syshyperblocks; j++)
                if (ftl_is_good_block(bbt, j) != 0)
                    if (ftl_vfl_read_page(i, j, 0, ftl_buffer,
                                          &ftl_sparebuffer) == 0)
                    {
                        struct ftl_vfl_cxt_type* cxt;
                        cxt = (struct ftl_vfl_cxt_type*)ftl_buffer;
                        memcpy(vflcxtblock, &(*cxt).vflcxtblocks, 8);
                        minusn = 0xFFFFFFFF;
                        vflcxtidx = 4;
                        for (k = 0; k < 4; k++)
                            if (vflcxtblock[k] != 0xFFFF)
                                if (ftl_vfl_read_page(i, vflcxtblock[k], 0,
                                                      ftl_buffer,
                                                      &ftl_sparebuffer) == 0)
                                    if (ftl_sparebuffer.meta.usn > 0
                                     && ftl_sparebuffer.meta.usn <= minusn)
                                    {
                                        minusn = ftl_sparebuffer.meta.usn;
                                        vflcxtidx = k;
                                    }
                        if (vflcxtidx == 4) return 1;
                        last = 0;
                        uint32_t max = (*ftl_nand_type).pagesperblock;
                        for (k = 8; k < max; k += 8)
                        {
                            if (ftl_vfl_read_page(i, vflcxtblock[vflcxtidx],
                                                  k, ftl_buffer,
                                                  &ftl_sparebuffer) != 0)
                                break;
                            last = k;
                        }
                        if (ftl_vfl_read_page(i, vflcxtblock[vflcxtidx],
                                              last, ftl_buffer,
                                              &ftl_sparebuffer) != 0)
                            return 1;
                        memcpy(&ftl_vfl_cxt[i], ftl_buffer, 0x800);
                        if (ftl_vfl_verify_checksum(i) != 0) return 1;
                        if (ftl_vfl_usn < ftl_vfl_cxt[i].usn)
                            ftl_vfl_usn = ftl_vfl_cxt[i].usn;
                        break;
                    }
        }
        else return 1;
    }
    return 0;
}


/* Commits the location of the FTL context blocks
   to a semi-randomly chosen VFL context */
uint32_t ftl_store_ctrl_block_list(void)
{
    uint32_t i;
    for (i = 0; i < ftl_banks; i++)
        memcpy(ftl_vfl_cxt[i].ftlctrlblocks, ftl_cxt.ftlctrlblocks, 6);
    return ftl_vfl_commit_cxt(ftl_vfl_usn % ftl_banks);
}


/* Initializes and mounts the VFL. */
uint32_t vfl_init(void)
{
    uint32_t i;
    uint32_t result = 0;
    uint32_t foundsignature, founddevinfo, blockwiped, repaired, skip;
    if (nand_init()) return 1;
    ftl_banks = 0;
    for (i = 0; i < 4; i++)
        if (nand_get_device_type(i) != 0) ftl_banks = i + 1;
    ftl_nand_type = nand_get_device_type(0);
    foundsignature = 0;
    blockwiped = 1;
    for (i = 0; i < (*ftl_nand_type).pagesperblock; i++)
    {
        result = nand_read_page(0, i, ftl_buffer, (uint32_t*)0, 1, 1);
        if ((result & 0x11F) == 0)
        {
            blockwiped = 0;
            if (((uint32_t*)ftl_buffer)[0] != 0x41303034) continue;
            foundsignature = 1;
            break;
        }
        else if ((result & 2) != 2) blockwiped = 0;
    }
    founddevinfo = ftl_has_devinfo();

    repaired = 0;
    skip = 0;
    if (!founddevinfo) return 1;
    if (foundsignature && (result & 0x11F) != 0) return 1;

    if (ftl_vfl_open() == 0) return 0;

    return 1;
}


uint32_t ftl_erase_block(uint32_t block)
{
    uint32_t i, j;
    ftl_erasectr[block]++;
    block = block + (*ftl_nand_type).blocks
          - (*ftl_nand_type).userblocks - 0x17;
    if (block == 0 || block >= (*ftl_nand_type).blocks) return 1;
    for (i = 0; i < ftl_banks; i++)
    {
        uint32_t pblock = ftl_vfl_get_physical_block(i, block);
        uint32_t rc;
        for (j = 0; j < 3; j++)
        {
            rc = nand_block_erase(i, pblock * (*ftl_nand_type).pagesperblock);
            if (rc == 0) break;
        }
        if (rc != 0) return 1;
    }
    return 0;
}


void bail(void)
{
    rendertext(&lcdbuffer[0x4fc1], 0, 0xffffffff, "nandfsck can't fix this yet.");
    rendertext(&lcdbuffer[0x5541], 0, 0xffffffff, "run diskmode to recover.");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    while(1);
}


uint32_t ftl_alloc_block()
{
    unsigned char lcdtext[30];
    while (1)
    {
        for (; firstfree < (*ftl_nand_type).userblocks + 0x17; firstfree++)
            if (!blk_type[firstfree]) break;
            else if (allocmode && blk_type[firstfree] != 1)
            {
                if (ftl_erase_block(firstfree))
                {
                    memset(lcdbuffer, 0xff, 0xb580);
                    rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, "Couldn't erase vBlock ");
                    snprintf(lcdtext, 30, "%d (pool alloc)!", firstfree);
                    rendertext(&lcdbuffer[0x631], 0, 0xffffffff, lcdtext);
                    bail();
                }
                break;
            }
        if (firstfree < (*ftl_nand_type).userblocks + 0x17)
        {
            blk_type[firstfree] = 1;
            return firstfree++;
        }
        if (!allocmode)
        {
            allocmode = 1;
            firstfree = 0;
        }
        else
        {
            memset(lcdbuffer, 0xff, 0xb580);
            rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, "Out of empty blocks!");
            bail();
        }
    }
}


void statusbar_init(uint32_t row, uint32_t range)
{
    if (!range) return;
    statusbar_y = row;
    statusbar_divisor = (range + 127) >> 7;
    uint32_t width;
    if (statusbar_divisor < 2)
    {
        statusbar_divisor = 0;
        statusbar_multiplier = 128 / (range + 1);
        width = (range + 1) * statusbar_multiplier;
    }
    else width = (range + 1) / statusbar_divisor;
    statusbar_offset = (176 - width) >> 1;
    displaylcd(statusbar_offset - 1, statusbar_offset - 1, row, row + 7, 0x40000000, 0);
    displaylcd(statusbar_offset + width + 1, statusbar_offset + width + 1,
               row, row + 7, 0x40000000, 0);
    displaylcd(statusbar_offset - 1, statusbar_offset + width + 1, row, row, 0x40000000, 0);
    displaylcd(statusbar_offset - 1, statusbar_offset + width + 1,
               row + 7, row + 7, 0x40000000, 0);
    statusbar_current = -1;
}


void statusbar_update(uint32_t pos)
{
    uint32_t x;
    if (!statusbar_divisor) x = pos * statusbar_multiplier;
    else x = pos / statusbar_divisor;
    if (statusbar_current == x) return;
    displaylcd(statusbar_offset + statusbar_current + 1, statusbar_offset + x,
               statusbar_y + 1, statusbar_y + 6, 0x40000000, 0x18);
    statusbar_current = x;
}


void main(void)
{
    uint32_t i, j, k;
    unsigned char lcdtext[30];

    memset(lcdbuffer, 0xff, 0xb580);
    rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, "nandfsck v0.1.1");
    rendertext(&lcdbuffer[0x631], 0, 0xffffffff, "Copyright 2010 TheSeven");
    rendertext(&lcdbuffer[0xbb1], 0, 0xffffffff, "Initializing...");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    if (vfl_init())
    {
        memset(lcdbuffer, 0xff, 0xb580);
        rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, "Could not mount VFL!");
        bail();
    }
    rendertext(&lcdbuffer[0x1131], 0, 0xffffffff, "Scanning flash...");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    statusbar_init(33, (*ftl_nand_type).userblocks + 0x17);
    uint32_t ppb = (*ftl_nand_type).pagesperblock * ftl_banks;
    memset(&ftl_cxt, 0x00, 0x800);
    memset(ftl_map, 0xff, 0x4000);
    memset(blk_usn, 0x00, 0x8000);
    memset(blk_type, 0x00, 0x2000);
    memset(ftl_erasectr, 0x00, 0x4000);
    memset(erasectr_usn, 0xff, 32);
    user_usn = 0;
    meta_usn = 0xffffffff;
    for (i = 0; i < (*ftl_nand_type).userblocks + 0x17; i++)
    {
        uint32_t ret = ftl_vfl_read((i + 1) * ppb - 1, 0, &ftl_sparebuffer, 1, 0);
        if ((ret & 0x11F) == 0 && ftl_sparebuffer.meta.type == 0x41)
        {
            uint32_t lbn = ftl_sparebuffer.user.lpn / ppb;
            if (ftl_sparebuffer.user.usn > user_usn)
                user_usn = ftl_sparebuffer.user.usn;
            if (ftl_sparebuffer.user.usn > blk_usn[lbn])
            {
                if (ftl_map[lbn] != 0xffff)
                    blk_type[ftl_map[lbn]] = 5;
                blk_usn[lbn] = ftl_sparebuffer.user.usn;
                ftl_map[lbn] = i;
                blk_type[i] = 1;
            }
            else blk_type[i] = 5;
        }
        else
            for (j = 0; j < ppb; j++)
            {
                ret = ftl_vfl_read(i * ppb + j, 0, &ftl_sparebuffer, 1, 0);
                if (ret & 2) break;
                if (ret & 0x11F)
                {
                    blk_type[i] = 4;
                    continue;
                }
                if (ftl_sparebuffer.meta.type == 0x40)
                {
                    blk_type[i] = 2;
                    break;
                }
                else if (ftl_sparebuffer.meta.type - 0x43 <= 4)
                {
                    blk_type[i] = 3;
                    if (ftl_sparebuffer.meta.type == 0x46)
                    {
                        uint32_t idx = ftl_sparebuffer.meta.idx;
                        if (ftl_sparebuffer.meta.usn < meta_usn)
                            meta_usn = ftl_sparebuffer.meta.usn;
                        if (ftl_sparebuffer.meta.usn < erasectr_usn[idx])
                        {
                            erasectr_usn[idx] = ftl_sparebuffer.meta.usn;
                            ret = ftl_vfl_read(i * ppb + j, &ftl_erasectr[idx << 10],
                                               &ftl_sparebuffer, 1, 0);
                            if (ret & 0x11f) memset(&ftl_erasectr[idx << 10], 0, 0x800);
                        }
                    }
                }
                else
                {
                    memset(lcdbuffer, 0xff, 0xb580);
                    snprintf(lcdtext, 30, "Invalid block type %02X while",
                             ftl_sparebuffer.meta.type);
                    rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, lcdtext);
                    snprintf(lcdtext, 30, "reading vPage %d", i * ppb + j);
                    rendertext(&lcdbuffer[0x631], 0, 0xffffffff, lcdtext);
                    bail();
                }
            }
        statusbar_update(i + 1);
    }

    uint32_t sum = 0;
    uint32_t count = 0;
    for (i = 0; i < 0x2000; i++)
        if (ftl_erasectr[i])
        {
            sum += ftl_erasectr[i];
            count++;
        }
    uint32_t average = sum / count;
    for (i = 0; i < 0x2000; i++)
        if (!ftl_erasectr[i])
            ftl_erasectr[i] = average;

    rendertext(&lcdbuffer[0x16b1], 0, 0xffffffff, "Committing scattered pages...");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    count = 0;
    for (i = 0; i < (*ftl_nand_type).userblocks + 0x17; i++)
        if (blk_type[i] == 2) count++;
    uint32_t block;
    uint32_t dirty;
    statusbar_init(41, count);
    if (count)
    {
        count = 0;
        for (i = 0; i < (*ftl_nand_type).userblocks + 0x17; i++)
            if (blk_type[i] == 2)
            {
                block = 0xffff;
                for (j = 0; j < (*ftl_nand_type).pagesperblock * ftl_banks; j++)
                {
                    uint32_t ret = ftl_vfl_read(i * ppb + j, ftl_buffer, &ftl_sparebuffer, 1, 0);
                    if (ret & 0x11F) continue;
                    if (ftl_sparebuffer.user.type != 0x40)
                    {
                        memset(lcdbuffer, 0xff, 0xb580);
                        snprintf(lcdtext, 30, "Invalid block type %02X while",
                                 ftl_sparebuffer.meta.type);
                        rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, lcdtext);
                        snprintf(lcdtext, 30, "reading vPage %d", i * ppb + j);
                        rendertext(&lcdbuffer[0x631], 0, 0xffffffff, lcdtext);
                        rendertext(&lcdbuffer[0xbb1], 0, 0xffffffff, "(scattered page)!");
                        bail();
                    }
                    if (block == 0xffff)
                    {
                        block = ftl_sparebuffer.user.lpn / ppb;
                        memset(pageusn, 0x00, 0x800);
                        memset(pagedata, 0x00, 0x100000);
                        if (ftl_map[block] != 0xffff)
                            for (k = 0; k < ppb; k++)
                            {
                                uint32_t ret = ftl_vfl_read(ftl_map[block] * ppb + k, pagedata[k],
                                                            &ftl_sparebuffer2, 1, 0);
                                if (ret & 0x11F) continue;
                                if (ftl_sparebuffer2.user.type != 0x40
                                 && ftl_sparebuffer2.user.type != 0x41)
                                {
                                    memset(lcdbuffer, 0xff, 0xb580);
                                    snprintf(lcdtext, 30, "Invalid block type %02X while",
                                             ftl_sparebuffer2.meta.type);
                                    rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, lcdtext);
                                    snprintf(lcdtext, 30, "reading vPage %d",
                                             ftl_map[block] * ppb + k);
                                    rendertext(&lcdbuffer[0x631], 0, 0xffffffff, lcdtext);
                                    rendertext(&lcdbuffer[0xbb1], 0, 0xffffffff,
                                               "(scattered page orig)!");
                                    bail();
                                }
                                if (block != ftl_sparebuffer2.user.lpn / ppb)
                                {
                                    memset(lcdbuffer, 0xff, 0xb580);
                                    rendertext(&lcdbuffer[0xb1], 0, 0xffffffff,
                                               "Foreign page in scattered");
                                    rendertext(&lcdbuffer[0x631], 0, 0xffffffff,
                                               "page orig block (vPage");
                                    snprintf(lcdtext, 30, "%d, LPN %d)!\n", ftl_map[block] * ppb + k,
                                             ftl_sparebuffer.user.usn);
                                    rendertext(&lcdbuffer[0xbb1], 0, 0xffffffff, lcdtext);
                                    bail();
                                }
                                pageusn[k] = ftl_sparebuffer2.user.usn;
                            }
                        dirty = 0;
                    }
                    if (block != ftl_sparebuffer.user.lpn / ppb)
                    {
                        memset(lcdbuffer, 0xff, 0xb580);
                        rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, "Foreign page in scattered");
                        rendertext(&lcdbuffer[0x631], 0, 0xffffffff, "page block (vPage");
                        snprintf(lcdtext, 30, "%d, LPN %d)!\n", i * ppb + j,
                                 ftl_sparebuffer.user.lpn);
                        rendertext(&lcdbuffer[0xbb1], 0, 0xffffffff, lcdtext);
                        bail();
                    }
                    uint32_t idx = ftl_sparebuffer.user.lpn % ppb;
                    if (ftl_sparebuffer.user.usn > user_usn)
                        user_usn = ftl_sparebuffer.user.usn;
                    if (ftl_sparebuffer.user.usn > pageusn[idx])
                    {
                        pageusn[idx] = ftl_sparebuffer.user.usn;
                        memcpy(pagedata[idx], ftl_buffer, 0x800);
                        dirty = 1;
                    }
                }
                if (dirty)
                {
                    if (ftl_erase_block(i))
                    {
                        memset(lcdbuffer, 0xff, 0xb580);
                        snprintf(lcdtext, 30, "Couldn't erase vBlock %d ", i);
                        rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, lcdtext);
                        rendertext(&lcdbuffer[0x631], 0, 0xffffffff, "(scattered page commit)!");
                        bail();
                    }
                    for (j = 0; j < ppb; j++)
                    {
                        memset(&ftl_sparebuffer, 0xFF, 0x40);
                        ftl_sparebuffer.user.lpn = block * ppb + j;
                        ftl_sparebuffer.user.usn = pageusn[j];
                        ftl_sparebuffer.user.type = 0x40;
                        if (j == ppb - 1) ftl_sparebuffer.user.type = 0x41;
                        if (ftl_vfl_write(i * ppb + j, pagedata[j], &ftl_sparebuffer))
                        {
                            memset(lcdbuffer, 0xff, 0xb580);
                            snprintf(lcdtext, 30, "Couldn't write vPage %d ", i * ppb + j);
                            rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, lcdtext);
                            rendertext(&lcdbuffer[0x631], 0, 0xffffffff,
                                       "(scattered page commit)!");
                            bail();
                        }
                    }
                    if (ftl_map[block] != 0xffff) blk_type[ftl_map[block]] = 5;
                    blk_type[i] = 1;
                    ftl_map[block] = i;
                }
                else blk_type[i] = 5;
                statusbar_update(++count);
            }
    }

    rendertext(&lcdbuffer[0x1c31], 0, 0xffffffff, "Fixing block map...");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    allocmode = 0;
    firstfree = 0;
    for (i = 0; i < 3; i++) ftl_cxt.ftlctrlblocks[i] = ftl_alloc_block();
    for (i = 0; i < 20; i++) ftl_cxt.blockpool[i] = ftl_alloc_block();
    for (i = 0; i < (*ftl_nand_type).userblocks; i++)
        if (ftl_map[i] == 0xffff)
            ftl_map[i] = ftl_alloc_block();
    ftl_cxt.usn = meta_usn - 1;
    ftl_cxt.nextblockusn = user_usn + 1;
    ftl_cxt.freecount = 20;
    ftl_cxt.clean_flag = 1;

    rendertext(&lcdbuffer[0x21b1], 0, 0xffffffff, "Committing FTL context...");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    uint32_t blockmappages = (*ftl_nand_type).userblocks >> 10;
    if (((*ftl_nand_type).userblocks & 0x1FF) != 0) blockmappages++;
    uint32_t erasectrpages = ((*ftl_nand_type).userblocks + 23) >> 10;
    if ((((*ftl_nand_type).userblocks + 23) & 0x1FF) != 0) erasectrpages++;
    uint32_t page = ftl_cxt.ftlctrlblocks[0] * ppb;
    for (i = 0; i < erasectrpages; i++)
    {
        memset(&ftl_sparebuffer, 0xFF, 0x40);
        ftl_sparebuffer.meta.usn = ftl_cxt.usn;
        ftl_sparebuffer.meta.idx = i;
        ftl_sparebuffer.meta.type = 0x46;
        if (ftl_vfl_write(page, &ftl_erasectr[i << 10], &ftl_sparebuffer))
        {
            memset(lcdbuffer, 0xff, 0xb580);
            snprintf(lcdtext, 30, "Couldn't write vPage %d ", page);
            rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, lcdtext);
            rendertext(&lcdbuffer[0x631], 0, 0xffffffff, "(save erase counters)!");
            bail();
        }
        ftl_cxt.ftl_erasectr_pages[i] = page++;
    }
    for (i = 0; i < blockmappages; i++)
    {
        memset(&ftl_sparebuffer, 0xFF, 0x40);
        ftl_sparebuffer.meta.usn = ftl_cxt.usn;
        ftl_sparebuffer.meta.idx = i;
        ftl_sparebuffer.meta.type = 0x44;
        if (ftl_vfl_write(page, &ftl_map[i << 10], &ftl_sparebuffer))
        {
            memset(lcdbuffer, 0xff, 0xb580);
            snprintf(lcdtext, 30, "Couldn't write vPage %d ", page);
            rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, lcdtext);
            rendertext(&lcdbuffer[0x631], 0, 0xffffffff, "(save block map)!");
            bail();
        }
        ftl_cxt.ftl_map_pages[i] = page++;
    }
    ftl_cxt.ftlctrlpage = page;
    memset(&ftl_sparebuffer, 0xFF, 0x40);
    ftl_sparebuffer.meta.usn = ftl_cxt.usn;
    ftl_sparebuffer.meta.type = 0x43;
    if (ftl_vfl_write(page, &ftl_cxt, &ftl_sparebuffer))
    {
        memset(lcdbuffer, 0xff, 0xb580);
        snprintf(lcdtext, 30, "Couldn't write vPage %d ", page);
        rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, lcdtext);
        rendertext(&lcdbuffer[0x631], 0, 0xffffffff, "(save FTL context)!");
        bail();
    }
    ftl_store_ctrl_block_list();
}


void irqhandler(void) __attribute__((interrupt("IRQ")));
void irqhandler(void)
{
    uint32_t sources;
    sources = SRCPND;

    SRCPND = sources;
    INTPND = sources;
}
